from typing import Literal, List, Optional
from pydantic import BaseModel
import urllib.request
import urllib.parse
import json
import yaml
from pathlib import Path
import xml.etree.ElementTree as ET



OSM_ENDPOINT = 'http://overpass-api.de/api/'
LOCATION_PREFIX = Path('src')
PARAMETERS_DIR = LOCATION_PREFIX / 'parameters'
AMENITY_SELECTORS_PATH = PARAMETERS_DIR / 'amenity_selectors.yaml'


ElementTypes = List[Literal['way', 'node', 'relation']]



# Output to frontend
class Landmark(BaseModel) :
    """
    A class representing a landmark or point of interest (POI) in the context of a trip.

    The Landmark class is used to model visitable locations, such as tourist attractions,
    natural sites, shopping locations, and start/end points in travel itineraries. It
    holds information about the landmark's attributes and supports comparisons and
    calculations, such as distance between landmarks.

    Attributes:
        name (str): The name of the landmark.
        type (Literal): The type of the landmark, which can be one of ['sightseeing', 'nature', 
            'shopping', 'start', 'finish'].
        location (tuple): A tuple representing the (latitude, longitude) of the landmark.
        osm_type (str): The OpenStreetMap (OSM) type of the landmark.
        osm_id (int): The OpenStreetMap (OSM) ID of the landmark.
        attractiveness (int): A score representing the attractiveness of the landmark.
        n_tags (int): The number of tags associated with the landmark.
        image_url (Optional[str]): A URL to an image of the landmark.
        website_url (Optional[str]): A URL to the landmark's official website.
        description (Optional[str]): A text description of the landmark.
        duration (Optional[int]): The estimated time to visit the landmark (in minutes).
        name_en (Optional[str]): The English name of the landmark.
        uuid (UUID): A unique identifier for the landmark, generated by default using uuid4.
        must_do (Optional[bool]): Whether the landmark is a "must-do" attraction.
        must_avoid (Optional[bool]): Whether the landmark should be avoided.
        is_secondary (Optional[bool]): Whether the landmark is secondary or less important.
        time_to_reach_next (Optional[int]): Estimated time (in minutes) to reach the next landmark.
        next_uuid (Optional[UUID]): UUID of the next landmark in sequence (if applicable).
    """

    # Properties of the landmark
    name : str
    type: Literal['sightseeing', 'nature', 'shopping', 'start', 'finish']
    location : tuple
    osm_type : str
    osm_id : int
    attractiveness : int
    n_tags : int
    image_url : Optional[str] = None
    website_url : Optional[str] = None
    wiki_url : Optional[str] = None
    description : Optional[str] = None                          # TODO future
    duration : Optional[int] = 0
    name_en : Optional[str] = None


    # Additional properties depending on specific tour
    must_do : Optional[bool] = False
    must_avoid : Optional[bool] = False
    is_secondary : Optional[bool] = False

    time_to_reach_next : Optional[int] = 0
    
    # More properties to define the score
    is_viewpoint : Optional[bool] = False
    is_cathedral : Optional[bool] = False
    is_place_of_worship : Optional[bool] = False


def OverpassQueryBuilder(area: tuple, element_types: ElementTypes, selector: str, 
                         conditions=[], out='center'):

    if not isinstance(conditions, list) :
        conditions = [conditions]

    query = '('
    search_area = f"({', '.join(map(str, area))})"
    if conditions :
        conditions = '(if: ' + ' && '.join(conditions) + ')'
    else :
        conditions = ''

    for elem in element_types :
        query += elem + '[' + selector + ']' + conditions + search_area + ';'

    query += ');' + f'out {out};'

    return query

def dict_to_selector_list(d: dict) -> list:
    """
    Convert a dictionary of key-value pairs to a list of Overpass query strings.

    Args:
        d (dict): A dictionary of key-value pairs representing the selector.

    Returns:
        list: A list of strings representing the Overpass query selectors.
    """
    return_list = []
    for key, value in d.items():
        if isinstance(value, list):
            val = '|'.join(value)
            return_list.append(f'{key}~"^({val})$"')
        elif isinstance(value, str) and len(value) == 0:
            return_list.append(f'{key}')
        else:
            return_list.append(f'{key}={value}')
    return return_list


def send_overpass_query(query: str) -> dict:
    """
    Sends the Overpass QL query to the Overpass API and returns the parsed JSON response.

    Args:
        query (str): The Overpass QL query to be sent to the Overpass API.

    Returns:
        dict: The parsed JSON response from the Overpass API, or None if the request fails.
    """

    # Define the Overpass API endpoint
    overpass_url = "https://overpass-api.de/api/interpreter"

    # Prepare the data to be sent as POST request, encoded as bytes
    data = urllib.parse.urlencode({'data': query}).encode('utf-8')

    # Create a custom header with a User-Agent
    headers = {
        'User-Agent': 'Mozilla/5.0 (compatible; OverpassQuery/1.0; +http://example.com)',
    }

    try:
        # Create a Request object with the specified URL, data, and headers
        request = urllib.request.Request(overpass_url, data=data, headers=headers)
        
        # Send the request and read the response
        with urllib.request.urlopen(request) as response:
            # Read and decode the response
            response_data = response.read().decode('utf-8')
            return ET.fromstring(response_data)

    except urllib.error.URLError as e:
        print(f"Error connecting to Overpass API: {e}")
        return None
    except json.JSONDecodeError:
        print("Error decoding the JSON response from Overpass API.")
        return None



def parse_result(root: ET.Element, elem_type) -> List[Landmark]:

    landmarks = []
    if root is None :
        return landmarks

    for osm_type in ['node', 'way', 'relation'] :
        for elem in root.findall(osm_type):

            # Extract basic info from the landmark.
            name = elem.find("tag[@k='name']").get('v') if elem.find("tag[@k='name']") is not None else None
            center = elem.find('center')
            tags = elem.findall('tag')

            # Extract the center latitude and longitude if available.
            if name is not None and center is not None:
                lat = center.get('lat')
                lon = center.get('lon')
                coords = tuple((lat, lon))
            else :
                continue
            
            # Convert this to Landmark object
            landmark = Landmark(name=name,
                                type=elem_type,
                                location=coords,
                                osm_id=elem.get('id'), 
                                osm_type=osm_type,
                                attractiveness=0,
                                n_tags=len(tags))

            # Browse through tags to add information to landmark.
            for tag in tags:
                key = tag.get('k')
                value = tag.get('v')

                # Skip this landmark if not suitable.
                if key == 'building:part' and value == 'yes' :
                    break
                if 'disused:' in key :
                    break
                if 'boundary:' in key :
                    break
                if 'shop' in key and elem_type != 'shopping' :
                    break
                # if value == 'apartments' :
                #     break
                
                # Fill in the other attributes.
                if key == 'image' :
                    landmark.image_url = value
                if key == 'website' :
                    landmark.website_url = value
                if key == 'place_of_worship' :
                    landmark.is_place_of_worship = True
                if key == 'wikipedia' :
                    landmark.wiki_url = value
                if key == 'name:en' :
                    landmark.name_en = value
                if 'building:' in key or 'pay' in key :
                    landmark.n_tags -= 1
                
                # Set the duration.
                if value in ['museum', 'aquarium', 'planetarium'] :
                    landmark.duration = 60
                elif value == 'viewpoint' :
                    landmark.is_viewpoint = True
                    landmark.duration = 10
                elif value == 'cathedral' :
                    landmark.is_place_of_worship = False
                    landmark.duration = 10
                else :
                    landmark.duration = 5

            else: 
                set_score(landmark, elem_type)
                landmarks.append(landmark)
            continue

    return landmarks


def set_score(landmark: Landmark, landmarktype: str) :

        score = landmark.n_tags**1.15
        if landmark.wiki_url :
            score *= 1.1
        if landmark.image_url :
            score *= 1.1
        if landmark.website_url :
            score *= 1.1
        if landmark.is_place_of_worship :
            score *= 0.65
        if landmark.is_viewpoint :
            # print(f"{landmark.name}:  n_tags={landmark.n_tags} and score={score*3*1.35*10}")
            score *= 3
        if landmarktype == 'nature' :
            score *= 1.35

        landmark.attractiveness = int(score * 10)


with AMENITY_SELECTORS_PATH.open('r') as f:
    amenity_selectors = yaml.safe_load(f)
    amenity_selector = amenity_selectors['nature']
bbox = tuple(('around:1714', 45.7576485, 4.8330241))

landmarks = []
for sel in dict_to_selector_list(amenity_selector):

    query = OverpassQueryBuilder(area=bbox,
                                 element_types=['way', 'relation'],
                                 selector=sel,
                                #  conditions='count_tags()>5',
                                 out='center')
    print(query + '\n')

    root = send_overpass_query(query)

    landmarks += parse_result(root, 'nature')


print(len(landmarks))