"""Module for finding public toilets around given coordinates."""
import logging
from OSMPythonTools.overpass import Overpass, overpassQueryBuilder
from OSMPythonTools.cachingStrategy import CachingStrategy, JSON

from ..structs.landmark import Toilets
from ..constants import OSM_CACHE_DIR


# silence the overpass logger
logging.getLogger('OSMPythonTools').setLevel(level=logging.CRITICAL)

class ToiletsManager:
    """
    Manages the process of fetching and caching toilet information from 
    OpenStreetMap (OSM) based on a specified location and radius.

    This class is responsible for:
    - Fetching toilet data from OSM using Overpass API around a given set of 
      coordinates (latitude, longitude).
    - Using a caching strategy to optimize requests by saving and retrieving 
      data from a local cache.
    - Logging important events and errors related to data fetching.

    Attributes:
        logger (logging.Logger): Logger for the class to capture events.
        location (tuple[float, float]): Latitude and longitude representing the 
            location to search around.
        radius (int): The search radius in meters for finding nearby toilets.
        overpass (Overpass): The Overpass API instance used to query OSM.
    """
    logger = logging.getLogger(__name__)

    location: tuple[float, float]
    radius: int    # radius in meters


    def __init__(self, location: tuple[float, float], radius : int) -> None:

        self.radius = radius
        self.location = location
        self.overpass = Overpass()
        CachingStrategy.use(JSON, cacheDir=OSM_CACHE_DIR)


    def generate_toilet_list(self) -> list[Toilets] :
        """
        Generates a list of toilet locations by fetching data from OpenStreetMap (OSM) 
        around the given coordinates stored in `self.location`.

        Returns:
        list[Toilets]: A list of `Toilets` objects containing detailed information 
                       about the toilets found around the given coordinates.
        """
        bbox = tuple((f"around:{self.radius}", str(self.location[0]), str(self.location[1])))
        toilets_list = []

        query = overpassQueryBuilder(
                bbox = bbox,
                elementType = ['node', 'way', 'relation'],
                # selector can in principle be a list already,
                # but it generates the intersection of the queries
                # we want the union
                selector = ['"amenity"="toilets"'],
                includeCenter = True,
                out = 'center'
                )
        self.logger.debug(f"Query: {query}")

        try:
            result = self.overpass.query(query)
        except Exception as e:
            self.logger.error(f"Error fetching landmarks: {e}")
            return None

        for elem in result.elements():
            location = (elem.centerLat(), elem.centerLon())

            # handle unprecise and no-name locations
            if location[0] is None:
                location = (elem.lat(), elem.lon())
            else :
                continue

            toilets = Toilets(location=location)

            if 'wheelchair' in elem.tags().keys() and elem.tag('wheelchair') == 'yes':
                toilets.wheelchair = True

            if 'changing_table' in elem.tags().keys() and elem.tag('changing_table') == 'yes':
                toilets.changing_table = True

            if 'fee' in elem.tags().keys() and elem.tag('fee') == 'yes':
                toilets.fee = True

            if 'opening_hours' in elem.tags().keys() :
                toilets.opening_hours = elem.tag('opening_hours')

            toilets_list.append(toilets)

        return toilets_list
